/* global IMWS, IMatch, $, MapModals */

var app = (() => {
  /////////////////////////////////////////////////////////////////////////////////////////////////
  // START of user-specific customizing section. Please adapt to your need!
  /////////////////////////////////////////////////////////////////////////////////////////////////
  // The name of the main Attribute Set
  const ATTR_SET_CALC = 'Calculation';
  // The name of the referenced global set containing customer data.
  // ATTENTION: To not overcomplicate matters, we assume there is only ONE referenced
  // attribute set, and this is a global set.
  const ATTR_SET_REF = 'Customer';
  // The attribute (of type Reference) in the Calculation Set that links to the Customer set
  const ATTR_LINK_FROM = 'refCustomerID';

  // These are your Attributes. You have to provide one entry for each attribute you want to use.
  // Throughout this app, the property name (e.g. QUANT) is used as key for this attribute.
  // For each entry, provide a configuration object with:
  //  attr: Name of the attribute. Be careful to spell it EXACTLY as in IMatch!
  //  id: The id you used in index.html for the corresponding form field.
  //  decimals: (optional, for attributes of numerical type only) Format the value to .
  //  refset: (for reference attributes only): The name of the referenced attribute set
  const ATTR = {
    DESCR: { attr: 'Description', id: 'attr_descr' },
    QUANT: { attr: 'Quantity', id: 'attr_quant' },
    PRICE: { attr: 'Price', id: 'attr_price', decimals: 2 },
    TAX: { attr: 'Tax Rate', id: 'attr_tax', decimals: 2 },
    TAX_AMT: { attr: 'Tax Amount', id: 'attr_tax_amt', decimals: 2 },
    TOTAL: { attr: 'Total', id: 'attr_total', decimals: 2 },
    GRAND_TOTAL: { attr: 'Grand Total', id: 'attr_grand_total', decimals: 2 },
    CUST_ID: { attr: 'refCustomerID', id: 'attr_cust_id', refset: 'Customer' }
  };
  // Now configure the attributes you need from the referenced master data set ATTR_SET_REF.
  // Make sure the property names are DIFFERENT to those from the set above. E.g. use a common
  // prefix CUST (for customer)
  const ATTR_REF = {
    CUSTNAME: { attr: 'Name', id: 'cust_name' },
    CUSTCOMPANY: { attr: 'Company', id: 'cust_company' },
    CUSTCITY: { attr: 'City', id: 'cust_city' }
  };

  // Workaround for writing back numerical values for locales using a comma as decimal separator:
  // Set to true if numbers should use a "," when POSTing attribute values
  const USE_DECIMAL_COMMA = false;

  // END of user-specific customizing section
  /////////////////////////////////////////////////////////////////////////////////////////////////

  // Global variables used in this app.
  var g = {
    // Here we will store the ID of the focused file
    fileId: 0,
    file: {}, // result from v1/files for the focused file
    instanceId: 0,
    mapCustomers: new Map(),
    hasAttributeRecord: false,
    schemaInfo: {},
    linkedCustomerId: 0
  };

  /**
   * USER SPECIFIC: CALCULATION LOGIC
   * This function is called whenever one of the input fields changes. It performs
   * the calculation for all derived fields and updates the UI accordingly.
   * If you add or change attributes, check if the calculation logic has to be adapted.
   */
  function performCalculation() {
    let vQuant = parseInt(_getFormVal(ATTR.QUANT));
    vQuant = isNaN(vQuant) ? 0 : vQuant;
    let vPrice = parseFloat(_getFormVal(ATTR.PRICE).replace(/,/, '.'));
    let vTaxPct = parseFloat(_getFormVal(ATTR.TAX).replace(/,/, '.'));
    let vValues = {};

    vValues[ATTR.TOTAL.attr] = vQuant * vPrice;
    vValues[ATTR.TAX_AMT.attr] = vTaxPct * vValues[ATTR.TOTAL.attr] / 100;
    vValues[ATTR.GRAND_TOTAL.attr] =
      vValues[ATTR.TOTAL.attr] + vValues[ATTR.TAX_AMT.attr];

    // Create an array of all calculated fields. Use
    let vFieldsToUpdate = [ATTR.TOTAL, ATTR.TAX_AMT, ATTR.GRAND_TOTAL];
    displayAttributeValues(vValues, vFieldsToUpdate);

    // Just a nice visual effect (from jQuery UI): Flash all fields that have been updated
    vFieldsToUpdate.forEach(element => {
      $('#' + element.id).effect('highlight', {}, 1000);
    });
  }

  /**
   * USER SPECIFIC: Displays some info for the focused file. At the moment, we show the
   * file name and the thumbnail only.
   * @param {File} pFileInfo 
   */
  function displayFileInfo(pFileInfo) {
    // Display file information
    $('#f_name').html(pFileInfo.name);
    //Display the thumbnail
    loadImage(pFileInfo.id, pFileInfo.protected);
  }

  /**
   * USER SPECIFIC: Displays some customer information. We assume that customers reside
   * in a global attribute set and are linked to the calculation set via a reference attribute.
   * @param id - The id of the customer record in the global attribute set
   */
  function displayCustomer(id) {
    // Get the specified customer
    let vCust = g.mapCustomers.get(Number(id));
    if (id != 0 && vCust) {
      // The '||' is a small trick: if an attribute in the Customer set is not filled,
      // the vCust[...] expression return undefined which is not valid to set via html().
      // Because undefined is a falsy value, we can use logical OR to provide an empty
      // string in this case.
      $('#' + ATTR_REF.CUSTNAME.id).html(vCust[ATTR_REF.CUSTNAME.attr] || '');
      $('#' + ATTR_REF.CUSTCOMPANY.id).html(
        vCust[ATTR_REF.CUSTCOMPANY.attr] || ''
      );
      $('#' + ATTR_REF.CUSTCITY.id).html(vCust[ATTR_REF.CUSTCITY.attr] || '');
    } else {
      // If no customer data is available, clear the fields
      $('#' + ATTR_REF.CUSTNAME.id).html('');
      $('#' + ATTR_REF.CUSTCOMPANY.id).html('');
      $('#' + ATTR_REF.CUSTCITY.id).html('');
    }

    //ADDED for V1.1: Drop Down list to display/select the customer
    // $('#CUST_' + id).attr('selected', 1);
    $('#cust-list option[value="' + id + '"]').attr('selected', true);
  }

  /**
   * Reads the attribute values of the focused file, displays them on the screen and
   * calculates sum fields etc.
   * This is the main function called whenever the focused file changes.
   */
  function readAndDisplayAttributes() {
    // Added for V1.2: If an Message Box is open, close it
    $('#imm-modal').modal('hide');

    // Get the needed fields for the focused file
    IMWS.get('v1/files', {
      idlist: IMatch.idlist.fileWindowFocusedFile,
      fields: 'id,name,datetime,protected'
    })
      .then(response => {
        // We expect exactly one record in the files array: the focused file
        if (response.files.length != 1) {
          g.fileId = undefined;
          g.hasAttributeRecord = false;
          $('#input-table').hide();
          return;
        } else {
          $('#input-table').show();
        }

        // Save the ID of the focusedFile. We need to know where the data came from
        // when we save the input later.
        g.fileId = response.files[0].id;
        g.file = response.files[0];

        displayFileInfo(response.files[0]);

        // IMPORTANT: Return the promise for the next step in chain to execute!
        return response;
      })
      .then(() => {
        return IMWS.get('v1/attributes', {
          set: ATTR_SET_CALC,
          id: g.fileId,
          // Do NOT specify as true! We need the internal id of the reference.
          resolverefs: false
        });
      })
      .then(response => {
        // We have to know if we have a record
        if (response.result == undefined || response.result.length != 1) {
          g.hasAttributeRecord = false;
          g.instanceId = 0;
        } else {
          g.hasAttributeRecord = true;
        }

        let vFile = response.result[0];

        let vAttr = vFile && vFile.data[0];
        if (g.hasAttributeRecord) {
          // data is an array of attribute records, each record with a distinct instanceid
          // WE SUPPORT ONLY ONE ATTRIBUTE RECORD PER FILE!
          if (vFile.data.length > 1) {
            $('#input-table').hide();
            errorMutipleRecords(g.file.name);
            return;
          }
          g.instanceId = vAttr.instanceId;
        }

        // Display the attribute values
        // _setFormVal([ATTR.QUANT, ATTR.PRICE, ATTR.DESCR, ATTR.TAX], vAttr);
        displayAttributeValues(vAttr);

        // Perform calculation. Adapt this function to your needs!
        performCalculation();

        // Display customer data. Save the internal id to the referenced customer for later.
        g.linkedCustomerId = vAttr ? vAttr[ATTR_LINK_FROM] : '0';
        // V1.2: Show the customer in the drop down list
        // First, reset the 'selected' flag for all entries
        $('#cust-list').find('option:selected').removeAttr('selected');
        // Second, set the chosen customer entry to 'selected'. ID=0 is reserved for the '(none)' entry
        // (meaning: no customer selected yet)
        $('#cust-list [value="' + g.linkedCustomerId + '"]').attr(
          'selected',
          true
        );
        displayCustomer(g.linkedCustomerId);
      });
  }

  /**
   * Copies the values from the attrValArray to the screen elements.
   * @param {array} attrValArray
   *@param {array} attrArray - (optional) An array of attributes. If not provided, all attributes configured in ATTR are used.
   */
  function displayAttributeValues(attrValArray, attrArray) {
    if (attrArray == undefined) {
      attrArray = Object.keys(ATTR).map(e => {
        return ATTR[e];
      });
    }

    attrArray.forEach(e => {
      if (e.decimals != undefined) {
        let vNum = Number(attrValArray ? attrValArray[e.attr] : 0);
        $('#' + e.id).val(vNum.toFixed(e.decimals));
      } else {
        // The id of the form element
        $('#' + e.id).val(attrValArray ? attrValArray[e.attr] : '');
      }
    }, this);
  }

  function _getFormVal(attr) {
    return $('#' + attr.id).val();
  }

  function errorMutipleRecords(pFileName) {
    let m = new MapModals();
    m.showMessageBoxOK(
      'Multiple attribute records',
      pFileName +
        ' has multiple attribute records.\n' +
        'This is not supported.',
      () => {
        console.log('YES clicked');
      },
      null
    );
  }

  // THIS FUNCTION HAS BEEN COPIED FROM THE FILEWINDOW SAMPLE APP.
  // This function sets the src attribute of the <img> tag to the url of the focused
  // file. It checks first if the file is protected.
  // If this is the case, it checks if the user has enabled the display of protected files.
  function loadImage(oid, isProtected) {
    // This is the url in the format we need for src= attributes.
    // Note that we need include the auto_token so IMatch knows we are allowed to access the file.
    // We load the 'T' size, which means thumbnail.
    // default=true means that the endpoint returns a standard 404 not found image if there is no image for the file.
    var url =
      IMatch.IMWSUrl() +
      '/v1/files/image?auth_token=' +
      IMatch.authToken() +
      '&id=' +
      oid +
      '&imagesize=T&default=true';

    // We can just show the file
    if (!isProtected) {
      var img = $('#f_img');
      img.attr('src', url);
      img.fadeIn();
    } else {
      // We need to check if protected files are enabled
      IMatch.showProtectedFiles().then(function(response) {
        // Construct the URL required to fetch the image.
        // We need to create the format expected by the src attribute of the <img> tag.
        if (response.protectedfilesVisible.value == true) {
          var img = $('#f_img');
          img.attr('src', url);
          img.fadeIn();
        } else {
          $('#f_img').hide();
        }
      });
    }
  }

  ///////////////////////////////////////////////////////////////////
  // Open the IMatch websocket and listen to events
  /* eslint-disable no-unused-vars */
  function _openWebSocket() {
    // If the app runs in IMatch, this is all we need.
    if (IMatch.isEmbedded()) {
      IMatch.openWebSocket();
    } else {
      // If this runs in a browser, we assume that IMatch listens
      // on the default IP address and port.
      IMatch.openWebSocket(IMatch.DEFAULT_WS_URL);
    }

    // React on events sent by IMatch.
    IMatch.webSocket().onmessage = function(e) {
      // Convert the message sent by IMatch to a proper JavaScript object:
      var msg = JSON.parse(e.data);

      // Process all events we're interested in.
      if (msg.msg == 'fileWindow.focusChanged') {
        // Selection in the file window has changed.
        app.render();
      }
    };

    // ADDED for V1.1: Also react on attribute events
    IMWS.openWebSocket();
    IMWS.webSocket().onmessage = function(e) {
      // Parse the message from string format into JSON.
      var msg = JSON.parse(e.data);

      if (msg.ma == IMWS.EVENT_CODES.evAttributeDB) {
        if (
          msg.mi == IMWS.EVENT_CODES.evSDBDataAdded ||
          msg.mi == IMWS.EVENT_CODES.evSDBDataUpdated ||
          msg.mi == IMWS.EVENT_CODES.evSDBDataDeleted
        ) {
          IMatch.debounce(app.render(), 500);
        }
      }
    };
  }
  /* eslint-enable no-unused-vars */

  /**
 * Initialize the App
 */
  function _init() {
    _openWebSocket();

    // Buffer the Customer global set
    IMWS.get('v1/attributes', {
      set: ATTR_SET_REF
    })
      .then(response => {
        // Read all customers into mapCustomers, indexed by the id internally used to link to a record.
        g.mapCustomers = new Map(response.data.map(el => [el.instanceId, el]));

        //<-- ADDED for V1.1: Drop down to select the customer
        // The id of each entry is CUST_ plus the instanceId of the linked customer record
        let list = $('#cust-list');
        list.text('');
        list.append('<option value="0" selected>(none)</option>');
        g.mapCustomers.forEach(cust => {
          list.append(
            '<option value="' + cust.instanceId + '">' + cust.Name + '</option>'
          );
        });
        //-->

        return response;
      })
      .then(() => {
        // Get information about the Customer attribute schema
        return IMWS.get('v1/attributes/schema', {
          set: ATTR_SET_CALC
        });
      })
      .then(response => {
        // Save the schema info. We will need it later for checking e.g. data types.
        g.schemaInfo = response.schema[0];
        return response;
      })
      .then(() => {
        // Now update the user interface
        readAndDisplayAttributes();
      });
  }

  /**
 * Called via Save button. Takes all relevant input fields and writes their values
 * into the corresponding attributes.
 */
  function _save() {
    // params
    var params = {
      // Which set to add records to?
      set: ATTR_SET_CALC,
      // ID of the file we got the data from
      id: g.fileId
    };

    // Now we setup the tasks array.
    var tasks = [];
    // One task will update the data for all attribute fields
    let task = { data: {} };
    // Fill the data array for all attributes
    task.data = fillTaskData();

    // If we did not get an instanceId, there is no attribute record yet.
    // In this case use 'add' to create one, otherwise 'update'.
    if (g.instanceId == 0) {
      task.op = 'add';
    } else {
      task.op = 'update';
      // ATTENTION: instanceid MUST be an array, even if you want to update only one instance.
      // That's the reason for putting [] around g.instanceId
      task.instanceid = [g.instanceId];
    }

    // Add task to thetasks array. In this case, there is only one task.
    tasks.push(task);
    // Convert data into a string to send it to IMWS
    // ATTENTION: parameter name has changed from actions to tasks (like in v1/metadata endpoints)
    // params.actions = JSON.stringify(actions);
    params.tasks = JSON.stringify(tasks);
    IMWS.post('v1/attributes', params)
      .then(response => {
        //
        console.log('Success: ' + JSON.stringify(response));
      })
      // Simple CatchAll. ToDo: Add more sophisticated error handling
      .catch(error => {
        console.log('Error! ' + JSON.stringify(error));
      }); //
  }

  /**
   * Transfers the content from the input fields into a data structure as needed by v1/attributes POST.
   * Depending on the data type, some conversions might be performed.
   *@return Object with attribute key/value pairs.
   */
  function fillTaskData() {
    let result = {};
    let vAttrArray = Object.keys(ATTR).map(function(e) {
      return [e, ATTR[e]];
    });
    vAttrArray.forEach(function(e) {
      // Check which data type the attribute has
      let vAttrInfo = g.schemaInfo.attributes.find(a => {
        return a.name == e[1].attr;
      });
      // Starting with V1.1, we can select the customer in the app
      // This is a special case because the id is stored in g.linkedCustomerId whenever the user
      // selects a different entry in the dop down list.
      if (
        vAttrInfo.dataType == 'reference' &&
        vAttrInfo.name == ATTR_LINK_FROM
      ) {
        result[e[1].attr] = g.linkedCustomerId;
      } else if (
        vAttrInfo.dataType == 'currency' ||
        vAttrInfo.dataType == 'real'
      ) {
        // Workaround: Values are not interpreted correctly by v1/attributes POST if
        // they use a point as decimal separator while the locale of the user is set
        // to using a comma. Therefore, replace . by , if needed.
        let vVal = $('#' + e[1].id).val();
        if (USE_DECIMAL_COMMA) {
          vVal = vVal.replace('.', ',');
        }
        result[e[1].attr] = vVal;
      } else {
        result[e[1].attr] = $('#' + e[1].id).val();
      }
    });

    return result;
  }

  /**
   * Called whenever the user selects a customer in the drop down box.
   * The value attribute of the selected entry contains the id of the
   * customer in the referenced global set. Store it and display the
   * new customer.
   */
  function _changeCustomer() {
    // let vTest = $('#cust-list').find('option:selected');
    let vNewCustomer = $('#cust-list').find('option:selected').attr('value');
    g.linkedCustomerId = vNewCustomer;
    displayCustomer(g.linkedCustomerId);
  }

  /**
   * This is only a test function to display something like a MsgBox.
   */
  function testModal() {
    // let m = new MapModals();
    // m.showMessageBoxOK('This is the title', 'and this the content');
    // $('#myMsgBoxOk').modal('show');

    let m = new MapModals();
    m.showMessageBoxYesNo(
      'This is the headline',
      'and this the message',
      () => {
        console.log('YES clicked');
      },
      null
    );
  }

  /*=== PUBLIC ===*/
  return {
    init: () => _init(),
    render: () => readAndDisplayAttributes(),
    recalculate: () => performCalculation(),

    btnSaveClicked: () => _save(),
    btnCancelClicked: () => readAndDisplayAttributes(),
    btnTestClicked: () => testModal(),

    // Added for V1.1
    changeCustomer: () => _changeCustomer()
  };
})();
